@@ -1,6 +1,11 @@
 """
 # Backtesting.py Documentation
 
+.. warning:: v0.2.0 breaking changes
+   Version 0.2.0 introduced some **breaking API changes**. For quick ways to
+   migrate existing 0.1.x code, see the implementing
+   [pull request](https://github.com/kernc/backtesting.py/pull/47/).
+
 ## Manuals
 
 * [**Quick Start User Guide**](../examples/Quick Start User Guide.html)
@@ -11,7 +16,7 @@
 * [Multiple Time Frames](../examples/Multiple Time Frames.html)
 * [Parameter Heatmap](../examples/Parameter Heatmap.html)
 
-These tutorials are also available as live notebooks:
+These tutorials are also available to test as live Jupyter notebooks:
 [![Binder](https://mybinder.org/badge_logo.svg)][binder]
 
 [binder]: \
@@ -22,6 +27,11 @@ urlpath=lab%2Ftree%2Fdoc%2Fexamples%2FQuick%20Start%20User%20Guide.ipynb
 
 * (contributions welcome)
 
+## FAQ
+
+Potentially outdated answers to popular questions can be found on the
+[issue tracker](https://github.com/kernc/backtesting.py/issues?q=label%3Aquestion).
+
 ## License
 
 This software is licensed under the terms of [AGPL 3.0]{: rel=license},
@@ -39,6 +49,6 @@ try:
 except ImportError:
     pass  # Package not installed
 
-from .backtesting import Backtest, Strategy, Orders, Position   # noqa: F401
+from .backtesting import Backtest, Strategy  # noqa: F401
 from . import lib  # noqa: F401
 from ._plotting import set_bokeh_output  # noqa: F401
