@@ -4,6 +4,7 @@ import sys
 import warnings
 from itertools import cycle, combinations
 from functools import partial
+from typing import List
 
 import numpy as np
 import pandas as pd
@@ -31,8 +32,7 @@ from bokeh.layouts import gridplot
 from bokeh.palettes import Category10
 from bokeh.transform import factor_cmap
 
-from backtesting._util import _data_period, _as_list
-
+from backtesting._util import _data_period, _as_list, _Indicator
 
 with open(os.path.join(os.path.dirname(__file__), 'autoscale_cb.js'),
           encoding='utf-8') as _f:
@@ -85,11 +85,73 @@ def lightness(color, lightness=.94):
     return color.to_rgb()
 
 
-def plot(*, results, df, indicators, filename='', plot_width=None,
+_MAX_CANDLES = 10000
+
+
+def _maybe_resample_data(resample_rule, df, indicators, equity_data, trades):
+    if isinstance(resample_rule, str):
+        freq = resample_rule
+    else:
+        if len(df) < _MAX_CANDLES:
+            return df, indicators, equity_data, trades
+
+        from_index = dict(day=-2, hour=-6, minute=1, second=0, millisecond=0,
+                          microsecond=0, nanosecond=0)[df.index.resolution]
+        FREQS = ('1T', '5T', '10T', '15T', '30T', '1H', '2H', '4H', '8H', '1D', '1W', '1M')
+        freq = next((f for f in FREQS[from_index:]
+                     if len(df.resample(f)) <= _MAX_CANDLES), FREQS[-1])
+        warnings.warn("Data contains too many candlesticks to plot; downsampling to {!r}. "
+                      "See `Backtest.plot(resample=...)`".format(freq))
+
+    from .lib import OHLCV_AGG, TRADES_AGG, _EQUITY_AGG
+    df = df.resample(freq, label='right').agg(OHLCV_AGG).dropna()
+
+    indicators = [_Indicator(i.df.resample(freq, label='right').mean()
+                             .dropna().reindex(df.index).values.T,
+                             **dict(i._opts, name=i.name,
+                                    # HACK: override `data` for its index
+                                    data=pd.Series(np.nan, index=df.index)))
+                  for i in indicators]
+    assert not indicators or indicators[0].df.index.equals(df.index)
+
+    equity_data = equity_data.resample(freq, label='right').agg(_EQUITY_AGG).dropna(how='all')
+    assert equity_data.index.equals(df.index)
+
+    def _weighted_returns(s, trades=trades):
+        df = trades.loc[s.index]
+        return ((df['Size'].abs() * df['ReturnPct']) / df['Size'].abs().sum()).sum()
+
+    def _group_trades(column):
+        def f(s, new_index=df.index.astype(np.int64), bars=trades[column]):
+            if s.size:
+                # Via int64 because on pandas recently broken datetime
+                mean_time = int(bars.loc[s.index].view('i8').mean())
+                new_bar_idx = new_index.get_loc(mean_time, method='nearest')
+                return new_bar_idx
+        return f
+
+    if len(trades):  # Avoid pandas "resampling on Int64 index" error
+        trades = trades.assign(count=1).resample(freq, on='ExitTime', label='right').agg(dict(
+            TRADES_AGG,
+            ReturnPct=_weighted_returns,
+            count='sum',
+            EntryBar=_group_trades('EntryTime'),
+            ExitBar=_group_trades('ExitTime'),
+        )).dropna()
+
+    return df, indicators, equity_data, trades
+
+
+def plot(*, results: pd.Series,
+         df: pd.DataFrame,
+         indicators: List[_Indicator],
+         filename='', plot_width=None,
          plot_equity=True, plot_pl=True,
          plot_volume=True, plot_drawdown=False,
-         smooth_equity=False, relative_equity=True, omit_missing=True,
-         superimpose=True, show_legend=True, open_browser=True):
+         smooth_equity=False, relative_equity=True,
+         superimpose=True, resample=True,
+         reverse_indicators=True,
+         show_legend=True, open_browser=True):
     """
     Like much of GUI code everywhere, this is a mess.
     """
@@ -101,39 +163,33 @@ def plot(*, results, df, indicators, filename='', plot_width=None,
     _bokeh_reset(filename)
 
     COLORS = [BEAR_COLOR, BULL_COLOR]
+    BAR_WIDTH = .8
 
-    orig_trade_data = trade_data = results._trade_data.copy(False)
+    assert df.index.equals(results['_equity_curve'].index)
+    equity_data = results['_equity_curve'].copy(deep=False)
+    trades = results['_trades']
 
-    orig_df = df = df.copy(False)
-    df.index.name = None  # Provides source name @index
-    index = df.index
-    time_resolution = getattr(index, 'resolution', None)
-    is_datetime_index = index.is_all_dates
-
-    # If all Volume is NaN, don't plot volume
     plot_volume = plot_volume and not df.Volume.isnull().all()
+    is_datetime_index = df.index.is_all_dates
 
-    # OHLC vbar width in msec.
-    # +1 will work in case of non-datetime index where vbar width should just be =1
-    bar_width = 1 + dict(day=86400,
-                         hour=3600,
-                         minute=60,
-                         second=1).get(time_resolution, 0) * 1000 * .85
+    from .lib import OHLCV_AGG
+    # ohlc df may contain many columns. We're only interested in, and pass on to Bokeh, these
+    df = df[list(OHLCV_AGG.keys())].copy(deep=False)
 
+    # Limit data to max_candles
     if is_datetime_index:
-        # Add index as a separate data source column because true .index is offset to align vbars
-        df['datetime'] = index
-        df.index = df.index + pd.Timedelta(bar_width / 2, unit='ms')
+        df, indicators, equity_data, trades = _maybe_resample_data(
+            resample, df, indicators, equity_data, trades)
 
-    if omit_missing:
-        bar_width = .8
-        df = df.reset_index(drop=True)
-        trade_data = trade_data.reset_index(drop=True)
-        index = df.index
+    df.index.name = None  # Provides source name @index
+    df['datetime'] = df.index  # Save original, maybe datetime index
+    df = df.reset_index(drop=True)
+    equity_data = equity_data.reset_index(drop=True)
+    index = df.index
 
     new_bokeh_figure = partial(
         _figure,
-        x_axis_type='datetime' if is_datetime_index and not omit_missing else 'linear',
+        x_axis_type='linear',
         plot_width=plot_width,
         plot_height=400,
         tools="xpan,xwheel_zoom,box_zoom,undo,redo,reset,crosshair,save",
@@ -150,21 +206,22 @@ def plot(*, results, df, indicators, filename='', plot_width=None,
 
     source = ColumnDataSource(df)
     source.add((df.Close >= df.Open).values.astype(np.uint8).astype(str), 'inc')
-    returns = trade_data['Returns'].dropna()
+
     trade_source = ColumnDataSource(dict(
-        index=returns.index,
-        datetime=orig_trade_data['Returns'].dropna().index,
-        exit_price=trade_data['Exit Price'].dropna(),
-        returns_pos=(returns > 0).astype(np.int8).astype(str),
+        index=trades['ExitBar'],
+        datetime=trades['ExitTime'],
+        exit_price=trades['ExitPrice'],
+        size=trades['Size'],
+        returns_positive=(trades['ReturnPct'] > 0).astype(int).astype(str),
     ))
 
     inc_cmap = factor_cmap('inc', COLORS, ['0', '1'])
-    cmap = factor_cmap('returns_pos', COLORS, ['0', '1'])
+    cmap = factor_cmap('returns_positive', COLORS, ['0', '1'])
     colors_darker = [lightness(BEAR_COLOR, .35),
                      lightness(BULL_COLOR, .35)]
-    trades_cmap = factor_cmap('returns_pos', colors_darker, ['0', '1'])
+    trades_cmap = factor_cmap('returns_positive', colors_darker, ['0', '1'])
 
-    if is_datetime_index and omit_missing:
+    if is_datetime_index:
         fig_ohlc.xaxis.formatter = FuncTickFormatter(
             args=dict(axis=fig_ohlc.xaxis[0],
                       formatter=DatetimeTickFormatter(days=['%d %b', '%a %d'],
@@ -178,7 +235,7 @@ return this.labels[index] || "";
         ''')
 
     NBSP = '&nbsp;' * 4
-    ohlc_extreme_values = df[['High', 'Low']].copy(False)
+    ohlc_extreme_values = df[['High', 'Low']].copy(deep=False)
     ohlc_tooltips = [
         ('x, y', NBSP.join(('$index',
                             '$y{0,0.0[0000]}'))),
@@ -216,36 +273,34 @@ return this.labels[index] || "";
     def _plot_equity_section():
         """Equity section"""
         # Max DD Dur. line
-        equity = trade_data['Equity']
-        argmax = trade_data['Drawdown Duration'].idxmax()
-        try:
-            dd_start = equity[:argmax].idxmax()
-        except Exception:  # ValueError: attempt to get argmax of an empty sequence
+        equity = equity_data['Equity'].copy()
+        dd_end = equity_data['DrawdownDuration'].idxmax()
+        if np.isnan(dd_end):
             dd_start = dd_end = equity.index[0]
-            timedelta = 0
         else:
-            dd_end = argmax
-            if is_datetime_index and omit_missing:
-                # "Calendar" duration
-                timedelta = df.datetime.iloc[dd_end] - df.datetime.iloc[dd_start]
-            else:
-                timedelta = dd_end - dd_start
-            # Get point intersection
-            if dd_end != index[-1]:
-                x1, x2 = index.get_loc(dd_end) - 1, index.get_loc(dd_end)
-                y, y1, y2 = equity[dd_start], equity[x1], equity[x2]
-                dd_end -= (1 - (y - y1) / (y2 - y1)) * (dd_end - index[x1])  # y = a x + b
+            dd_start = equity[:dd_end].idxmax()
+            # If DD not extending into the future, get exact point of intersection with equity
+            if dd_end != equity.index[-1]:
+                dd_end = np.interp(equity[dd_start],
+                                   (equity[dd_end - 1], equity[dd_end]),
+                                   (dd_end - 1, dd_end))
 
         if smooth_equity:
-            select = (trade_data[['Entry Price',
-                                  'Exit Price']].dropna(how='all').index |
-                      # Include beginning
-                      equity.index[:1] |
-                      # Include max dd end points. Otherwise, the MaxDD line looks amiss.
-                      pd.Index([dd_start, dd_end]))
-            equity = equity[select].reindex(equity.index)
+            interest_points = pd.Index([
+                # Beginning and end
+                equity.index[0], equity.index[-1],
+                # Peak equity and peak DD
+                equity.idxmax(), equity_data['DrawdownPct'].idxmax(),
+                # Include max dd end points. Otherwise the MaxDD line looks amiss.
+                dd_start, int(dd_end), min(int(dd_end + 1), equity.size - 1),
+            ])
+            select = pd.Index(trades['ExitBar']) | interest_points
+            select = select.unique().dropna()
+            equity = equity.iloc[select].reindex(equity.index)
             equity.interpolate(inplace=True)
 
+        assert equity.index.equals(equity_data.index)
+
         if relative_equity:
             equity /= equity.iloc[0]
 
@@ -287,14 +342,15 @@ return this.labels[index] || "";
                     color='blue', size=8)
 
         if not plot_drawdown:
-            drawdown = trade_data['Drawdown']
+            drawdown = equity_data['DrawdownPct']
             argmax = drawdown.idxmax()
             fig.scatter(argmax, equity[argmax],
                         legend_label='Max Drawdown (-{:.1f}%)'.format(100 * drawdown[argmax]),
                         color='red', size=8)
-        fig.line([dd_start, dd_end], equity[dd_start],
+        dd_timedelta_label = df['datetime'].iloc[int(round(dd_end))] - df['datetime'].iloc[dd_start]
+        fig.line([dd_start, dd_end], equity.iloc[dd_start],
                  line_color='red', line_width=2,
-                 legend_label='Max Dd Dur. ({})'.format(timedelta)
+                 legend_label='Max Dd Dur. ({})'.format(dd_timedelta_label)
                  .replace(' 00:00:00', '')
                  .replace('(0 days ', '('))
 
@@ -303,7 +359,7 @@ return this.labels[index] || "";
     def _plot_drawdown_section():
         """Drawdown section"""
         fig = new_indicator_figure(y_axis_label="Drawdown")
-        drawdown = trade_data['Drawdown']
+        drawdown = equity_data['DrawdownPct']
         argmax = drawdown.idxmax()
         source.add(drawdown, 'drawdown')
         r = fig.line('index', 'drawdown', source=source, line_width=1.3)
@@ -319,20 +375,26 @@ return this.labels[index] || "";
         fig = new_indicator_figure(y_axis_label="Profit / Loss")
         fig.add_layout(Span(location=0, dimension='width', line_color='#666666',
                             line_dash='dashed', line_width=1))
-        position = trade_data['Exit Position'].dropna()
-        returns_long = returns.copy()
-        returns_short = returns.copy()
-        returns_long[position < 0] = np.nan
-        returns_short[position > 0] = np.nan
+        returns_long = np.where(trades['Size'] > 0, trades['ReturnPct'], np.nan)
+        returns_short = np.where(trades['Size'] < 0, trades['ReturnPct'], np.nan)
+        size = trades['Size'].abs()
+        size = np.interp(size, (size.min(), size.max()), (8, 20))
         trade_source.add(returns_long, 'returns_long')
         trade_source.add(returns_short, 'returns_short')
-        MARKER_SIZE = 13
+        trade_source.add(size, 'marker_size')
+        if 'count' in trades:
+            trade_source.add(trades['count'], 'count')
         r1 = fig.scatter('index', 'returns_long', source=trade_source, fill_color=cmap,
-                         marker='triangle', line_color='black', size=MARKER_SIZE)
+                         marker='triangle', line_color='black', size='marker_size')
         r2 = fig.scatter('index', 'returns_short', source=trade_source, fill_color=cmap,
-                         marker='inverted_triangle', line_color='black', size=MARKER_SIZE)
-        set_tooltips(fig, [("P/L", "@returns_long{+0.[000]%}")], vline=False, renderers=[r1])
-        set_tooltips(fig, [("P/L", "@returns_short{+0.[000]%}")], vline=False, renderers=[r2])
+                         marker='inverted_triangle', line_color='black', size='marker_size')
+        tooltips = [("Size", "@size{0,0}")]
+        if 'count' in trades:
+            tooltips.append(("Count", "@count{0,0}"))
+        set_tooltips(fig, tooltips + [("P/L", "@returns_long{+0.[000]%}")],
+                     vline=False, renderers=[r1])
+        set_tooltips(fig, tooltips + [("P/L", "@returns_short{+0.[000]%}")],
+                     vline=False, renderers=[r2])
         fig.yaxis.formatter = NumeralTickFormatter(format="0.[00]%")
         return fig
 
@@ -342,15 +404,16 @@ return this.labels[index] || "";
         fig.xaxis.formatter = fig_ohlc.xaxis[0].formatter
         fig.xaxis.visible = True
         fig_ohlc.xaxis.visible = False  # Show only Volume's xaxis
-        r = fig.vbar('index', bar_width, 'Volume', source=source, color=inc_cmap)
+        r = fig.vbar('index', BAR_WIDTH, 'Volume', source=source, color=inc_cmap)
         set_tooltips(fig, [('Volume', '@Volume{0.00 a}')], renderers=[r])
         fig.yaxis.formatter = NumeralTickFormatter(format="0 a")
         return fig
 
     def _plot_superimposed_ohlc():
         """Superimposed, downsampled vbars"""
+        time_resolution = pd.DatetimeIndex(df['datetime']).resolution
         resample_rule = (superimpose if isinstance(superimpose, str) else
-                         dict(day='W',
+                         dict(day='M',
                               hour='D',
                               minute='H',
                               second='T',
@@ -362,13 +425,13 @@ return this.labels[index] || "";
                 stacklevel=4)
             return
 
-        orig_df['_width'] = 1
-        from .lib import OHLCV_AGG
-        df2 = orig_df.resample(resample_rule, label='left').agg(dict(OHLCV_AGG, _width='count'))
+        df2 = (df.assign(_width=1).set_index('datetime')
+               .resample(resample_rule, label='left')
+               .agg(dict(OHLCV_AGG, _width='count')))
 
         # Check if resampling was downsampling; error on upsampling
-        orig_freq = _data_period(orig_df)
-        resample_freq = _data_period(df2)
+        orig_freq = _data_period(df['datetime'])
+        resample_freq = _data_period(df2.index)
         if resample_freq < orig_freq:
             raise ValueError('Invalid value for `superimpose`: Upsampling not supported.')
         if resample_freq == orig_freq:
@@ -376,48 +439,33 @@ return this.labels[index] || "";
                           stacklevel=4)
             return
 
-        if omit_missing:
-            width2 = '_width'
-            df2.index = df2['_width'].cumsum().shift(1).fillna(0)
-            df2.index += df2['_width'] / 2 - .5
-            df2['_width'] -= .1  # Candles don't touch
-        else:
-            del df['_width']
-            width2 = dict(day=86400 * 5,
-                          hour=86400,
-                          minute=3600,
-                          second=60)[time_resolution] * 1000
-            df2.index += pd.Timedelta(
-                width2 / 2 +
-                (width2 / 5 if resample_rule == 'W' else 0),  # Sunday week start
-                unit='ms')
-        df2['inc'] = (df2.Close >= df2.Open).astype(np.uint8).astype(str)
+        df2.index = df2['_width'].cumsum().shift(1).fillna(0)
+        df2.index += df2['_width'] / 2 - .5
+        df2['_width'] -= .1  # Candles don't touch
+
+        df2['inc'] = (df2.Close >= df2.Open).astype(int).astype(str)
         df2.index.name = None
         source2 = ColumnDataSource(df2)
         fig_ohlc.segment('index', 'High', 'index', 'Low', source=source2, color='#bbbbbb')
         colors_lighter = [lightness(BEAR_COLOR, .92),
                           lightness(BULL_COLOR, .92)]
-        fig_ohlc.vbar('index', width2, 'Open', 'Close', source=source2, line_color=None,
+        fig_ohlc.vbar('index', '_width', 'Open', 'Close', source=source2, line_color=None,
                       fill_color=factor_cmap('inc', colors_lighter, ['0', '1']))
 
     def _plot_ohlc():
         """Main OHLC bars"""
         fig_ohlc.segment('index', 'High', 'index', 'Low', source=source, color="black")
-        r = fig_ohlc.vbar('index', bar_width, 'Open', 'Close', source=source,
+        r = fig_ohlc.vbar('index', BAR_WIDTH, 'Open', 'Close', source=source,
                           line_color="black", fill_color=inc_cmap)
         return r
 
     def _plot_ohlc_trades():
         """Trade entry / exit markers on OHLC plot"""
-        exit_price = trade_data['Exit Price'].dropna()
-        entry_price = trade_data['Entry Price'].dropna().iloc[:exit_price.size]  # entry can be one more at the end  # noqa: E501
-        trade_source.add(np.column_stack((entry_price.index, exit_price.index)).tolist(),
-                         'position_lines_xs')
-        trade_source.add(np.column_stack((entry_price, exit_price)).tolist(),
-                         'position_lines_ys')
+        trade_source.add(trades[['EntryBar', 'ExitBar']].values.tolist(), 'position_lines_xs')
+        trade_source.add(trades[['EntryPrice', 'ExitPrice']].values.tolist(), 'position_lines_ys')
         fig_ohlc.multi_line(xs='position_lines_xs', ys='position_lines_ys',
                             source=trade_source, line_color=trades_cmap,
-                            legend_label='Trades ({})'.format(len(trade_data)),
+                            legend_label='Trades ({})'.format(len(trades)),
                             line_width=8, line_alpha=1, line_dash='dotted')
 
     def _plot_indicators():
@@ -441,6 +489,7 @@ return this.labels[index] || "";
                 return self is other
 
         ohlc_colors = colorgen()
+        indicator_figs = []
 
         for i, value in enumerate(indicators):
             value = np.atleast_2d(value)
@@ -456,7 +505,7 @@ return this.labels[index] || "";
                 fig = fig_ohlc
             else:
                 fig = new_indicator_figure()
-                figs_below_ohlc.append(fig)
+                indicator_figs.append(fig)
             tooltips = []
             colors = value._opts['color']
             colors = colors and cycle(_as_list(colors)) or (
@@ -476,7 +525,7 @@ return this.labels[index] || "";
                             'index', source_name, source=source,
                             legend_label=legend_label, color=color,
                             line_color='black', fill_alpha=.8,
-                            marker='circle', radius=bar_width / 2 * 1.5)
+                            marker='circle', radius=BAR_WIDTH / 2 * 1.5)
                     else:
                         fig.line(
                             'index', source_name, source=source,
@@ -487,7 +536,7 @@ return this.labels[index] || "";
                         r = fig.scatter(
                             'index', source_name, source=source,
                             legend_label=LegendStr(legend_label), color=color,
-                            marker='circle', radius=bar_width / 2 * .9)
+                            marker='circle', radius=BAR_WIDTH / 2 * .9)
                     else:
                         r = fig.line(
                             'index', source_name, source=source,
@@ -509,6 +558,7 @@ return this.labels[index] || "";
                 # have the legend only contain text without the glyph
                 if len(value) == 1:
                     fig.legend.glyph_width = 0
+        return indicator_figs
 
     # Construct figure ...
 
@@ -530,7 +580,10 @@ return this.labels[index] || "";
 
     ohlc_bars = _plot_ohlc()
     _plot_ohlc_trades()
-    _plot_indicators()
+    indicator_figs = _plot_indicators()
+    if reverse_indicators:
+        indicator_figs = indicator_figs[::-1]
+    figs_below_ohlc.extend(indicator_figs)
 
     set_tooltips(fig_ohlc, ohlc_tooltips, vline=True, renderers=[ohlc_bars])
 
